/*============================================================================
  File:     The Tipping Point.sql

  Summary:  Various examples and ways to see the estimated and actual tipping
			point of when a nonclustered index is not used vs. doing a table
			scan.
  
  SQL Server Version: 2005+
------------------------------------------------------------------------------
  Written by Kimberly L. Tripp, SYSolutions, Inc.

  For more scripts and sample code, check out 
    http://www.SQLskills.com

  This script is intended only as a supplement to demos and lectures
  given by Kimberly L. Tripp.  
  
  THIS CODE AND INFORMATION ARE PROVIDED "AS IS" WITHOUT WARRANTY OF 
  ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED 
  TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
  PARTICULAR PURPOSE.
============================================================================*/

-- You can get AdventureWorks2012 here: http://msftdbprodsamples.codeplex.com/
-- Or, you can use this same script with AdventureWorks2008 or 2008R2 (same location)
-- Be sure to do the math based on the output of the version you're working with
-- this script has the numbers/math for AW 2012.

USE AdventureWorksDW2012
go

sp_helpindex  FactInternetSales

-- Get my version of sp_helpindex here: http://www.sqlskills.com/blogs/kimberly/use-this-sp_helpindex-rewrites/ 
exec sp_SQLskills_SQL2008_helpindex FactInternetSales
go

-- For 2000
-- DBCC SHOWCONTIG .. WITH TABLERESULTS

SELECT * FROM sys.dm_db_index_physical_stats
	(db_id(), object_id('FactInternetSales'), 
	    1, NULL, 'detailed')  -- for a big table limited or sampled

-- Clustered Index has 2048 PAGES (Level 0 in CL)
-- 1/4 = ~512
------------
-- 1/3 = ~610

SELECT S.CustomerKey, COUNT(*) AS NumTotalSales
FROM FactInternetSales AS S
GROUP BY S.CustomerKey
ORDER BY S.CustomerKey 
--COMPUTE AVG(COUNT(*))
go

-- Turn on showplan (Ctrl+K)
SET STATISTICS IO ON
SET STATISTICS TIME ON
go

dbcc dropcleanbuffers
go

SELECT C.CustomerKey, C.LastName, sum(S.SalesAmount)
FROM dbo.FactInternetSales AS S
	INNER JOIN dbo.DimCustomer AS C
		ON S.CustomerKey = C.CustomerKey
WHERE C.CustomerKey < 11087
GROUP BY C.CustomerKey, C.LastName
OPTION (MAXDOP 1)
go

dbcc dropcleanbuffers
go

SELECT C.CustomerKey, C.LastName, sum(S.SalesAmount)
FROM dbo.FactInternetSales AS S
	INNER JOIN dbo.DimCustomer AS C
		ON S.CustomerKey = C.CustomerKey
WHERE C.CustomerKey < 11088
GROUP BY C.CustomerKey, C.LastName
OPTION (MAXDOP 1)
go

-- Is this still the same in a bigger database on a beefy machine?
-- the concept is the same but the tipping point might be higher

-- Please note this next part is based on a modified version of 
-- AdventureWorksDW2008_ModifiedSalesKey
-- Updated/backed up June 2012

USE AdventureWorksDW2008_ModifiedSalesKey
go

SELECT * FROM sys.dm_db_index_physical_stats
	(db_id(), object_id('FactInternetSales'), 
	    1, NULL, 'detailed')  -- for a big table limited or sampled
-- 475754 pages
-- select 475754/4  --118938 (is 1/4 of the pages)
-- select 475754/3  --158584 (is 1/3 of the pages)

SET STATISTICS IO ON
go

DBCC DROPCLEANBUFFERS
go
SELECT C.CustomerKey, C.LastName, sum(S.SalesAmount)
FROM dbo.FactInternetSales AS S
	INNER JOIN dbo.DimCustomer AS C
		ON S.CustomerKey = C.CustomerKey
WHERE C.CustomerKey < 11042
GROUP BY C.CustomerKey, C.LastName
OPTION (MAXDOP 1)
go

DBCC DROPCLEANBUFFERS
go
SELECT C.CustomerKey, C.LastName, sum(S.SalesAmount)
FROM dbo.FactInternetSales AS S 
	INNER JOIN dbo.DimCustomer AS C
		ON S.CustomerKey = C.CustomerKey
WHERE C.CustomerKey < 11043
GROUP BY C.CustomerKey, C.LastName
OPTION (MAXDOP 1)
go

DBCC DROPCLEANBUFFERS
go
SELECT C.CustomerKey, C.LastName, sum(S.SalesAmount)
FROM dbo.FactInternetSales AS S WITH (FORCESEEK)
	INNER JOIN dbo.DimCustomer AS C
		ON S.CustomerKey = C.CustomerKey
WHERE C.CustomerKey < 11043
GROUP BY C.CustomerKey, C.LastName
OPTION (MAXDOP 1)
go

DBCC DROPCLEANBUFFERS
go
SELECT C.CustomerKey, C.LastName, sum(S.SalesAmount)
FROM dbo.FactInternetSales AS S 
	INNER JOIN dbo.DimCustomer AS C
		ON S.CustomerKey = C.CustomerKey
WHERE C.CustomerKey < 15000
GROUP BY C.CustomerKey, C.LastName
OPTION (MAXDOP 1)
go

DBCC DROPCLEANBUFFERS
go
SELECT C.CustomerKey, C.LastName, sum(S.SalesAmount)
FROM dbo.FactInternetSales AS S WITH (FORCESEEK)
	INNER JOIN dbo.DimCustomer AS C
		ON S.CustomerKey = C.CustomerKey
WHERE C.CustomerKey < 15000
GROUP BY C.CustomerKey, C.LastName
OPTION (MAXDOP 1)
go

-- What if we let SQL Server use multiple processors/cores

SELECT S.CustomerKey, C.LastName, sum(S.SalesAmount)
FROM dbo.FactInternetSales AS S
	INNER JOIN dbo.DimCustomer AS C
		ON S.CustomerKey = C.CustomerKey
WHERE S.CustomerKey < 11037
GROUP BY S.CustomerKey, C.LastName
go

SELECT S.CustomerKey, C.LastName, sum(S.SalesAmount)
FROM dbo.FactInternetSales AS S
	INNER JOIN dbo.DimCustomer AS C
		ON S.CustomerKey = C.CustomerKey
WHERE S.CustomerKey < 11038
GROUP BY S.CustomerKey, C.LastName
go

--SELECT S.CustomerKey, C.LastName, sum(S.SalesAmount)
--FROM dbo.FactInternetSales AS S WITH (FORCESEEK)
--	INNER JOIN dbo.DimCustomer AS C
--		ON S.CustomerKey = C.CustomerKey
--WHERE S.CustomerKey < 11063
--GROUP BY S.CustomerKey, C.LastName
go

-- If this query were critical... 
-- I'd first look at the missing index suggestion 
--   (the "green" suggestion in showplan)

--CREATE NONCLUSTERED INDEX [<Name of Missing Index, sysname,>]
--ON [dbo].[FactInternetSales] ([CustomerKey])
--INCLUDE ([SalesAmount])

-- But... before you create it make sure to FIRST check existing indexes:
-- You can grab sp_SQLskills_SQL2008_helpindex here: 

sp_SQLskills_SQL2008_helpindex FactInternetSales
go

--1) Create the new index

CREATE NONCLUSTERED INDEX IX_FactInternetSales_CustomerKey_INCSalesAmount
ON [dbo].[FactInternetSales] ([CustomerKey])
INCLUDE ([SalesAmount])
go

--2) Drop the old???
--   we should always check our indexes first to see if any can be dropped/consolidated 
--   based on the new index being created!

DROP INDEX [FactInternetSales].IX_FactInternetSales_CustomerKey
go

-- May need to drop the old before creating the new 
-- (if space is an issue)
-- but this may negatively affect existing users?!

-- Now, no matter what the range and/or number of rows, this index can always be used:
SELECT S.CustomerKey, C.LastName, sum(S.SalesAmount)
FROM dbo.FactInternetSales AS S
	INNER JOIN dbo.DimCustomer AS C
		ON S.CustomerKey = C.CustomerKey
WHERE S.CustomerKey < 11048
GROUP BY S.CustomerKey, C.LastName
go

SELECT S.CustomerKey, C.LastName, sum(S.SalesAmount)
FROM dbo.FactInternetSales AS S
	INNER JOIN dbo.DimCustomer AS C
		ON S.CustomerKey = C.CustomerKey
WHERE S.CustomerKey < 11049
GROUP BY S.CustomerKey, C.LastName
go

-- What about an even bigger range:
SELECT S.CustomerKey, C.LastName, sum(S.SalesAmount)
FROM dbo.FactInternetSales AS S 
	INNER JOIN dbo.DimCustomer AS C
		ON S.CustomerKey = C.CustomerKey
WHERE S.CustomerKey > 11049
GROUP BY S.CustomerKey, C.LastName
go

SELECT S.CustomerKey, C.LastName, sum(S.SalesAmount)
FROM dbo.FactInternetSales AS S WITH (FORCESEEK)
	INNER JOIN dbo.DimCustomer AS C
		ON S.CustomerKey = C.CustomerKey
WHERE S.CustomerKey > 11049
GROUP BY S.CustomerKey, C.LastName
go

------------------------------------------------------------------
-- Clean up and prep for next execution:
------------------------------------------------------------------
USE AdventureWorksDW2008_ModifiedSalesKey
go

CREATE NONCLUSTERED INDEX IX_FactInternetSales_CustomerKey
ON [dbo].[FactInternetSales] ([CustomerKey])
go

DROP INDEX [FactInternetSales].IX_FactInternetSales_CustomerKey_INCSalesAmount
go
